from pandas import DataFrame, read_excel
import pickle

from funki import λ
from funki.terminal import Terminal
from funki import report

class Funk:
  def __init__(
    self,
    name:str,
    io_value_pairs=[],
    output_properties=[],
    description='',
    setup=λ.default_setup,
    run=λ.nop_σ_x,
    tear_down=λ.nop_x,
  ):
    self.name = name
    self._state = {'terminal': Terminal()}
    self.description = description
    self._run = run
    self._tear_down = tear_down

    setup_result = setup()
    if setup_result is None:
      setup_result = {}

    for key in setup_result.keys():
      self._state[key] = setup_result[key]

    if not io_value_pairs and not output_properties:
      raise ValueError(
        'Must specify either io_value pairs or output properties.'
      )

    self.io_value_pairs = io_value_pairs
    self._output_properties = output_properties

  def run(self, x):
    z_return = self._run(self._state, x)
    if self._state['terminal'] == 'test':
      z_stdo = self._state['terminal'].output_stream_as_list
      self._state['terminal'].reset()
      return (z_return, z_stdo)
    else:
      return z_return

  def _load_xyz(self, io_pair):
    __input, __target = io_pair
    xyz ={}
    
    if 'x' not in __input.keys():
      __input['x'] = None
    xyz['x_args']   = __input['x']

    if 'stdo' not in __target.keys():
      __target['stdo'] = []
    xyz['y_stdo'] = __target['stdo']
    
    if 'y' not in __target.keys():
      __target['y'] = "DO_NOT_CARE"
    xyz['y_return'] = __target['y']

    xyz['z_return'] = self.run(xyz['x_args'])
    xyz['z_stdo'] = self._state['terminal'].output_stream_as_list
    xyz['x_stdi']  = __input['stdi']

    if 'file' in __target.keys():
      xyz['y_file'] = __target['file']['content']
      if 'read_as' in __target['file']:
        if __target['file']['read_as'] == 'xlsx as df':
          xyz['z_file'] = read_excel(__target['file']['filename'])
        elif __target['file']['read_as'] == 'pickle':
          with open(__target['file']['filename'], 'rb') as φ:
            xyz['z_file'] = pickle.load(φ)
        else:
          raise NotImplementedError('Need to write this branch')
      else:
        raise NotImplementedError('Need to write this branch')
    return xyz

  def observation_does_not_match_expectation(self, xyz):
    if any([
      isinstance(xyz['y_return'], DataFrame),
      isinstance(xyz['z_return'], DataFrame)
    ]):
      # Dealing with a DataFrame
      if not xyz['y_return'].equals(xyz['z_return']):
        print(xyz['y_return'].compare(xyz['z_return']))

      return any([
        not xyz['y_return'].equals(xyz['z_return']),
        xyz['z_stdo'] != xyz['y_stdo']
      ])
    else:
      # Not dealing with a DataFrame
      if xyz['y_return'] == "DO_NOT_CARE":
        return False # Ignore this case as it is marked as DO_NOT_CARE
      else:
        return any([
          xyz['z_return'] != xyz['y_return'],
          xyz['z_stdo'] != xyz['y_stdo']
        ])

  def _test(self):
    """
    Input x provided to function f returns obervation z
    x --> f(x) --> z

    Input x provided to function f is expected to return y
    x --> f(x) --> y

    Test passes is expected return value y is equal to observation z.

    x includes:
      x_args: arguments passed to the function as single value or dict of values
      x_stdi: stdi stream provided to the function while running
    
    y includes:
      y_return: the expected return value from calling f(x)
      y_stdo: the expected stdo data emitted by f(x) during it's execution.
    
    z includes:
      z_return: the observed return value from calling f(x)
      z_stdo: the observed stdo data emitted by f(x) during it's execution.
    """

    if len(self.io_value_pairs) <= 0:
      # Check that the output satisfies the output property constraints if
      # no input is supplied.
      if self._output_properties:
        z_return = self.run({})

      for output_property in self._output_properties:
        has_output_property = output_property(z_return)
        if not has_output_property:
          return report.property_failure(self.name, z_return)

    for _io_pair in self.io_value_pairs:
      self._state['terminal'].reset_stdo()

      __input, _ = _io_pair

      if 'stdi' not in __input.keys():
        __input['stdi'] = []

      if isinstance(__input['stdi'], list):
        self._state['terminal'].input_stream_as_list += __input['stdi']
      
      xyz = self._load_xyz(_io_pair)

      if self.observation_does_not_match_expectation(xyz):
        return report.fail(self.name, xyz)

      for output_property in self._output_properties:
        z_return = xyz['z_return']
        has_output_property = output_property(z_return)
        if not has_output_property:
          return report.property_failure(self.name, z_return)
    return report.success(self.name)

  def test(self):
    self._state['terminal'].mode = 'test'
    result = self._test()
    self._state['terminal'].mode = 'run'
    return result

  __del__ = lambda self: self._tear_down(self._state)
